@@ -23,6 +23,7 @@
 // Qt includes
 #include <QtCore/QByteArray>
 #include <QtCore/QStringList>
+#include <QtCore/QTemporaryFile>
 #include <QtDBus/QDBusConnection>
 #include <QtDBus/QDBusServiceWatcher>
 
@@ -30,10 +31,12 @@
 #include <apt-pkg/algorithms.h>
 #include <apt-pkg/depcache.h>
 #include <apt-pkg/error.h>
+#include <apt-pkg/fileutl.h>
 #include <apt-pkg/init.h>
 #include <apt-pkg/policy.h>
 #include <apt-pkg/sourcelist.h>
 #include <apt-pkg/strutl.h>
+#include <apt-pkg/tagfile.h>
 
 // Xapian includes
 #include <xapian.h>
@@ -41,6 +44,7 @@
 // QApt includes
 #include "cache.h"
 #include "config.h"
+#include "debfile.h"
 #include "workerdbus.h" // OrgKubuntuQaptworkerInterface
 
 namespace QApt {
@@ -48,12 +52,19 @@ namespace QApt {
 class BackendPrivate
 {
 public:
-    BackendPrivate() : cache(0), config(0), records(0), maxStackSize(20), state(InvalidEvent) {}
+    BackendPrivate()
+        : state(InvalidEvent)
+        , cache(0)
+        , records(0)
+        , maxStackSize(20)
+        , config(0)
+    {
+    }
     ~BackendPrivate()
     {
         delete cache;
-        delete config;
         delete records;
+        delete config;
     }
     // Caches
     // The canonical list of all unique, non-virutal package objects
@@ -189,7 +200,8 @@ void Backend::reloadCache()
 
         Package *pkg = new Package(this, depCache, d->records, iter);
         d->packagesIndex[iter->ID] = count;
-        d->packages.insert(count++, pkg);
+        d->packages.append(pkg);
+        ++count;
 
         if (iter->CurrentVer) {
             d->installedCount++;
@@ -216,6 +228,45 @@ void Backend::reloadCache()
 
     d->undoStack.clear();
     d->redoStack.clear();
+
+    // Determine which packages are pinned for display purposes
+    QString dir = QString::fromStdString(_config->FindDir("Dir::Etc"))
+                  % QLatin1String("preferences.d/");
+    QDir logDirectory(dir);
+    QStringList pinFiles = logDirectory.entryList(QDir::Files, QDir::Name);
+    pinFiles << QString::fromStdString(_config->FindDir("Dir::Etc")) %
+                QLatin1String("preferences");
+
+    Q_FOREACH (const QString &pinName, pinFiles) {
+        QString pinPath;
+
+        // Handle absolute paths
+        if (!pinName.startsWith(QLatin1Char('/'))) {
+            pinPath = dir % pinName;
+        } else {
+            pinPath = pinName;
+        }
+        QFile pinFile(pinPath);
+
+        if (!pinFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
+            continue;
+        }
+        FileFd Fd(pinFile.handle(), FileFd::ReadOnly);
+
+        pkgTagFile tagFile(&Fd);
+        if (_error->PendingError()) {
+            continue;
+        }
+
+        pkgTagSection tags;
+        while (tagFile.Step(tags)) {
+            QLatin1String name(tags.FindS("Package").c_str());
+            Package *pkg = package(name);
+            if (pkg) {
+                pkg->setPinned(true);
+            }
+        }
+    }
 }
 
 void Backend::throwInitError()
@@ -276,10 +327,6 @@ Package *Backend::packageForFile(const QString &file) const
 {
     Q_D(const Backend);
 
-    if (file.isEmpty()) {
-        return 0;
-    }
-
     Q_FOREACH (Package *package, d->packages) {
         if (package->installedFilesList().contains(file)) {
             return package;
@@ -363,12 +410,6 @@ qint64 Backend::downloadSize() const
     // Raw size, ignoring already-downloaded or partially downloaded archives
     qint64 downloadSize = d->cache->depCache()->DebSize();
 
-    // If downloadSize() is called during a cache refresh, there is a chance it
-    // will do so at a bad time and produce an error. Discard any errors that
-    // happen during this function, since they will always be innocuous and at
-    // worst will result in the less accurate DebSize() number being returned
-    _error->PushToStack();
-
     // If possible, get what really needs to be downloaded
     pkgAcquire fetcher;
     pkgPackageManager *PM = _system->CreatePM(d->cache->depCache());
@@ -377,9 +418,6 @@ qint64 Backend::downloadSize() const
     }
     delete PM;
 
-    _error->Discard();
-    _error->RevertToStack();
-
     return downloadSize;
 }
 
@@ -595,7 +633,7 @@ CacheState Backend::currentCacheState() const
     #if QT_VERSION >= 0x040700
     state.reserve(pkgSize);
     #endif
-    for (unsigned i = 0; i < pkgSize; ++i) {
+    for (int i = 0; i < pkgSize; ++i) {
         state.append(d->packages[i]->state());
     }
 
@@ -628,7 +666,7 @@ void Backend::restoreCacheState(const CacheState &state)
     pkgDepCache *deps = d->cache->depCache();
     pkgDepCache::ActionGroup group(*deps);
 
-    for (unsigned i = 0; i < d->packages.size(); ++i) {
+    for (int i = 0; i < d->packages.size(); ++i) {
         Package *pkg = d->packages[i];
         int flags = pkg->state();
         int oldflags = state[i];
@@ -796,8 +834,53 @@ void Backend::commitChanges()
     d->worker->commitChanges(packageList);
 }
 
+void Backend::downloadArchives(const QString &listFile, const QString &destination)
+{
+    Q_D(Backend);
+
+    QFile file(listFile);
+
+    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
+        // TODO: error
+        return;
+    }
+
+    QByteArray buffer = file.readAll();
+
+    QList<QByteArray> lines = buffer.split('\n');
+
+    if (lines.isEmpty() || lines.first() != QByteArray("[Download List]")) {
+        return;
+    }
+
+    lines.removeAt(0);
+
+    QStringList packages;
+    foreach (const QByteArray &line, lines) {
+        packages << line;
+    }
+
+    QString dirName = listFile.left(listFile.lastIndexOf('/'));
+
+    QDir dir(dirName);
+    dir.mkdir(QLatin1String("packages"));
+
+    d->worker->setLocale(QLatin1String(setlocale(LC_MESSAGES, 0)));
+    d->worker->downloadArchives(packages, destination);
+}
+
+void Backend::installDebFile(const DebFile &debFile)
+{
+    Q_D(Backend);
+
+    d->worker->setLocale(QLatin1String(setlocale(LC_MESSAGES, 0)));
+    d->worker->installDebFile(debFile.filePath());
+}
+
 void Backend::packageChanged(Package *package)
 {
+    Q_UNUSED(package);
+
     emit packageChanged();
 }
 
@@ -814,7 +897,7 @@ bool Backend::saveInstalledPackagesList(const QString &path) const
     Q_D(const Backend);
 
     QString selectionDocument;
-    for (unsigned i = 0; i < d->packages.size(); ++i) {
+    for (int i = 0; i < d->packages.size(); ++i) {
 
         if (d->packages.at(i)->isInstalled()) {
             selectionDocument.append(d->packages[i]->name() %
@@ -834,7 +917,7 @@ bool Backend::saveSelections(const QString &path) const
     Q_D(const Backend);
 
     QString selectionDocument;
-    for (unsigned i = 0; i < d->packages.size(); ++i) {
+    for (int i = 0; i < d->packages.size(); ++i) {
         int flags = d->packages.at(i)->state();
 
         if (flags & Package::ToInstall) {
@@ -940,6 +1023,122 @@ bool Backend::loadSelections(const QString &path)
     return true;
 }
 
+bool Backend::saveDownloadList(const QString &path) const
+{
+    Q_D(const Backend);
+
+    QString downloadDocument;
+    downloadDocument.append(QLatin1String("[Download List]") % QLatin1Char('\n'));
+    for (int i = 0; i < d->packages.size(); ++i) {
+        int flags = d->packages.at(i)->state();
+
+        if (flags & Package::ToInstall) {
+            downloadDocument.append(d->packages[i]->name() % QLatin1Char('\n'));
+        }
+    }
+
+    return d->writeSelectionFile(downloadDocument, path);
+}
+
+bool Backend::setPackagePinned(Package *package, bool pin)
+{
+    Q_D(Backend);
+
+    QString dir = QString::fromStdString(_config->FindDir("Dir::Etc"))
+                  % QLatin1String("preferences.d/");
+    QString path = dir % package->latin1Name();
+    QString pinDocument;
+
+    if (pin) {
+        if (package->state() & Package::IsPinned) {
+            return true;
+        }
+
+        pinDocument = QLatin1Literal("Package: ") % package->latin1Name()
+                      % QLatin1Char('\n');
+
+        if (package->installedVersion().isEmpty()) {
+            pinDocument += QLatin1String("Pin: version  0.0\n");
+        } else {
+            pinDocument += QLatin1Literal("Pin: version ") % package->installedVersion()
+                           % QLatin1Char('\n');
+        }
+
+        // Make configurable?
+        pinDocument += QLatin1String("Pin-Priority: 1001\n\n");
+    } else {
+        QDir logDirectory(dir);
+        QStringList pinFiles = logDirectory.entryList(QDir::Files, QDir::Name);
+        pinFiles << QString::fromStdString(_config->FindDir("Dir::Etc")) %
+                    QLatin1String("preferences");
+
+        // Search all pin files, delete package stanza from file
+        Q_FOREACH (const QString &pinName, pinFiles) {
+            QString pinPath;
+            if (!pinName.startsWith(QLatin1Char('/'))) {
+                pinPath = dir % pinName;
+            } else {
+                pinPath = pinName;
+            }
+            QFile pinFile(pinPath);
+
+            // Open to get a file name
+            QTemporaryFile tempFile;
+            if (!tempFile.open()) {
+                return false;
+            }
+            tempFile.close();
+
+            QString tempFileName = tempFile.fileName();
+            FILE *out = fopen(tempFileName.toStdString().c_str(),"w");
+            if (!out) {
+                return false;
+            }
+
+            if (!pinFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
+                continue;
+            }
+            FileFd Fd(pinFile.handle(), FileFd::ReadOnly);
+
+            pkgTagFile tagFile(&Fd);
+            if (_error->PendingError()) {
+                return false;
+            }
+
+            pkgTagSection tags;
+            while (tagFile.Step(tags)) {
+                QString name = QLatin1String(tags.FindS("Package").c_str());
+
+                if (name.isEmpty()) {
+                    return false;
+                }
+
+                // Include all but the matching name in the new pinfile
+                if (name != package->latin1Name()) {
+                    TFRewriteData tfrd;
+                    tfrd.Tag = 0;
+                    tfrd.Rewrite = 0;
+                    tfrd.NewTag = 0;
+                    TFRewrite(out, tags, TFRewritePackageOrder, &tfrd);
+                    fprintf(out, "\n");
+                }
+            }
+
+            if (!tempFile.open()) {
+                return false;
+            }
+
+            pinDocument = tempFile.readAll();
+        }
+    }
+
+    if (!d->worker->writeFileToDisk(pinDocument, path)) {
+        return false;
+    }
+
+    return true;
+}
+
 void Backend::updateXapianIndex()
 {
     Q_D(Backend);
@@ -947,6 +1146,43 @@ void Backend::updateXapianIndex()
     d->worker->updateXapianIndex();
 }
 
+bool Backend::addArchiveToCache(const DebFile &archive)
+{
+    Q_D(Backend);
+
+    // Sanity checks
+    Package *pkg = package(archive.packageName());
+    if (!pkg) {
+        // The package is not in the cache, so we can't do anything
+        // with this .deb
+        return false;
+    }
+
+    QString arch = archive.architecture();
+
+    if (arch != QLatin1String("all") &&
+        arch != d->config->readEntry(QLatin1String("APT::Architecture"), QString())) {
+        // Incompatible architecture
+        return false;
+    }
+
+    QString debVersion = archive.version();
+    QString candVersion = pkg->availableVersion();
+
+    if (debVersion != candVersion) {
+        // Incompatible version
+        return false;
+    }
+
+    if (archive.md5Sum() != pkg->md5Sum()) {
+        // Not the same as the candidate
+        return false;
+    }
+
+    // Add the package, but we'll need auth so the worker'll do it
+    return d->worker->copyArchiveToCache(archive.filePath());
+}
+
 void Backend::workerStarted()
 {
     Q_D(Backend);
@@ -962,6 +1198,8 @@ void Backend::workerStarted()
             this, SIGNAL(downloadMessage(int, const QString&)));
     connect(d->worker, SIGNAL(commitProgress(const QString&, int)),
             this, SIGNAL(commitProgress(const QString&, int)));
+    connect(d->worker, SIGNAL(debInstallMessage(const QString&)),
+            this, SIGNAL(debInstallMessage(const QString&)));
     connect(d->worker, SIGNAL(questionOccurred(int, const QVariantMap&)),
             this, SLOT(emitWorkerQuestionOccurred(int, const QVariantMap&)));
     connect(d->worker, SIGNAL(warningOccurred(int, const QVariantMap&)),
@@ -972,6 +1210,8 @@ void Backend::workerFinished(bool result)
 {
     Q_D(Backend);
 
+    Q_UNUSED(result);
+
     d->state = InvalidEvent;
 
     disconnect(d->watcher, SIGNAL(serviceOwnerChanged(QString,QString,QString)),
@@ -985,6 +1225,8 @@ void Backend::workerFinished(bool result)
                this, SIGNAL(downloadMessage(int, const QString&)));
     disconnect(d->worker, SIGNAL(commitProgress(const QString&, int)),
                this, SIGNAL(commitProgress(const QString&, int)));
+    disconnect(d->worker, SIGNAL(debInstallMessage(const QString&)),
+               this, SIGNAL(debInstallMessage(const QString&)));
     disconnect(d->worker, SIGNAL(questionOccurred(int, const QVariantMap&)),
                this, SLOT(emitWorkerQuestionOccurred(int, const QVariantMap&)));
     disconnect(d->worker, SIGNAL(warningOccurred(int, const QVariantMap&)),
@@ -1030,8 +1272,10 @@ void Backend::emitWorkerQuestionOccurred(int question, const QVariantMap &detail
 
 void Backend::serviceOwnerChanged(const QString &name, const QString &oldOwner, const QString &newOwner)
 {
+    Q_UNUSED(name);
+
     if (oldOwner.isEmpty()) {
-        return; // Don't care, just appearing
+        return; // Don't care if empty, just appearing
     }
 
     if (newOwner.isEmpty()) {
