@@ -31,6 +31,7 @@
 
 // Apt includes
 #include <apt-pkg/algorithms.h>
+#include <apt-pkg/debversion.h>
 #include <apt-pkg/depcache.h>
 #include <apt-pkg/indexfile.h>
 #include <apt-pkg/init.h>
@@ -51,15 +52,27 @@ namespace QApt {
 class PackagePrivate
 {
     public:
-        ~PackagePrivate() { delete packageIter; };
+        PackagePrivate()
+            : state(0)
+            , staticStateCalculated(false)
+        {
+        }
+        ~PackagePrivate()
+        {
+            delete packageIter;
+        };
         QApt::Backend *backend;
         pkgDepCache *depCache;
         pkgRecords *records;
         pkgCache::PkgIterator *packageIter;
         int state;
+        bool staticStateCalculated;
 
         pkgCache::PkgFileIterator searchPkgFileIter(const QLatin1String &label, const QString &release) const;
         QString getReleaseFileForOrigin(const QLatin1String &label, const QString &release) const;
+
+        // Calculate state flags that cannot change
+        void initStaticState(const pkgCache::VerIterator &ver, pkgDepCache::StateCache &stateCache);
 };
 
 pkgCache::PkgFileIterator PackagePrivate::searchPkgFileIter(const QLatin1String &label, const QString &release) const
@@ -105,13 +118,12 @@ QString PackagePrivate::getReleaseFileForOrigin(const QLatin1String &label, cons
         for(I=list->begin(); I != list->end(); ++I) {
             vector<pkgIndexFile *>  *ifv = (*I)->GetIndexFiles();
             if(find(ifv->begin(), ifv->end(), index) != ifv->end()) {
-                string stduri = _config->FindDir("Dir::State::lists");
-                stduri += URItoFileName((*I)->GetURI());
-                stduri += "dists_";
-                stduri += (*I)->GetDist();
-                stduri += "_Release";
+                QString uri = QString::fromStdString(_config->FindDir("Dir::State::lists"))
+                % QString::fromStdString(URItoFileName((*I)->GetURI()))
+                % QLatin1Literal("dists_")
+                % QString::fromStdString((*I)->GetDist())
+                % QLatin1Literal("_Release");
 
-                QString uri = QLatin1String(stduri.c_str());
                 return uri;
             }
         }
@@ -120,6 +132,70 @@ QString PackagePrivate::getReleaseFileForOrigin(const QLatin1String &label, cons
     return QString();
 }
 
+void PackagePrivate::initStaticState(const pkgCache::VerIterator &ver, pkgDepCache::StateCache &stateCache)
+{
+    int packageState = 0;
+
+    if (!ver.end()) {
+        packageState |= QApt::Package::Installed;
+
+        if (stateCache.CandidateVer && stateCache.Upgradable()) {
+            packageState |= QApt::Package::Upgradeable;
+            if (stateCache.Keep()) {
+                packageState |= QApt::Package::Held;
+            }
+        }
+
+        if (stateCache.Downgrade()) {
+            packageState |= QApt::Package::ToDowngrade;
+        }
+    } else {
+        packageState |= QApt::Package::NotInstalled;
+    }
+    if (stateCache.NowBroken()) {
+        packageState |= QApt::Package::NowBroken;
+    }
+
+    if (stateCache.InstBroken()) {
+        packageState |= QApt::Package::InstallBroken;
+    }
+
+    if ((*packageIter)->Flags & (pkgCache::Flag::Important |
+                                 pkgCache::Flag::Essential)) {
+        packageState |= QApt::Package::IsImportant;
+    }
+
+    if ((*packageIter)->CurrentState == pkgCache::State::ConfigFiles) {
+        packageState |= QApt::Package::ResidualConfig;
+    }
+
+    if (!stateCache.CandidateVer) {
+        packageState |= QApt::Package::NotDownloadable;
+    } else if (!stateCache.CandidateVerIter(*depCache).Downloadable()) {
+        packageState |= QApt::Package::NotDownloadable;
+    }
+
+    if (stateCache.Flags & pkgCache::Flag::Auto) {
+        packageState |= QApt::Package::IsAuto;
+    }
+
+    if (stateCache.Garbage) {
+        packageState |= QApt::Package::IsGarbage;
+    }
+
+    if (stateCache.NowPolicyBroken()) {
+        packageState |= QApt::Package::NowPolicyBroken;
+    }
+
+    if (stateCache.InstPolicyBroken()) {
+        packageState |= QApt::Package::InstallPolicyBroken;
+    }
+
+    state |= packageState;
+
+    staticStateCalculated = true;
+}
+
 Package::Package(QApt::Backend* backend, pkgDepCache *depCache,
                  pkgRecords *records, pkgCache::PkgIterator &packageIter)
         : d(new PackagePrivate())
@@ -217,8 +293,7 @@ QString Package::longDescription() const
         // Split at double newline, by "section"
         QStringList sections = rawDescription.split(QLatin1String("\n ."));
 
-        int i;
-        for (i = 0; i < sections.count(); ++i) {
+        for (int i = 0; i < sections.count(); ++i) {
             sections[i].replace(QRegExp(QLatin1String("\n( |\t)+(-|\\*)")),
                                 QLatin1Literal("\n\r ") % QString::fromUtf8("\xE2\x80\xA2"));
             // There should be no new lines within a section.
@@ -279,6 +354,29 @@ QString Package::version() const
     }
 }
 
+QString Package::upstreamVersion() const
+{
+    const char *ver;
+
+    if (!(*d->packageIter)->CurrentVer) {
+        pkgDepCache::StateCache &State = (*d->depCache)[*d->packageIter];
+        if (!State.CandidateVer) {
+            return QString();
+        } else {
+            ver = State.CandidateVerIter(*d->depCache).VerStr();
+        }
+    } else {
+        ver = d->packageIter->CurrentVer().VerStr();
+    }
+
+    return QString::fromStdString(_system->VS->UpstreamVersion(ver));
+}
+
+QString Package::upstreamVersion(const QString &version)
+{
+    return QString::fromStdString(_system->VS->UpstreamVersion(version.toStdString().c_str()));
+}
+
 QStringList Package::availableVersions() const
 {
     QStringList versions;
@@ -392,6 +490,23 @@ QString Package::component() const
     return QLatin1String(File.Component());
 }
 
+QByteArray Package::md5Sum() const
+{
+    QByteArray md5Sum;
+
+    const pkgCache::VerIterator &ver = (*d->depCache).GetCandidateVer(*d->packageIter);
+
+    if(ver.end()) {
+        return md5Sum;
+    }
+
+    pkgRecords::Parser &rec = d->records->Lookup(ver.FileList());
+    md5Sum = rec.MD5Hash().c_str();
+
+    return md5Sum;
+
+}
+
 QUrl Package::changelogUrl() const
 {
     QUrl url;
@@ -491,9 +606,12 @@ QString Package::supportedUntil() const
         return QString();
     }
 
-    QString supportTimeString = controlField(QLatin1String("Supported"));
-    if (supportTimeString.isEmpty()) {
-        return QString();
+    // Default to 18m in case the package has no "supported" field
+    QString supportTimeString = QLatin1String("18m");
+    QString supportTimeField = controlField(QLatin1String("Supported"));
+
+    if (!supportTimeString.isEmpty()) {
+        supportTimeString = supportTimeField;
     }
 
     QChar unit = supportTimeString.at(supportTimeString.length() - 1);
@@ -502,9 +620,9 @@ QString Package::supportedUntil() const
 
     QDateTime supportEnd;
 
-    if (unit == 'm') {
+    if (unit == QLatin1Char('m')) {
         supportEnd = QDateTime::fromTime_t(releaseDate).addMonths(supportTime);
-    } else if (unit == 'y') {
+    } else if (unit == QLatin1Char('y')) {
         supportEnd = QDateTime::fromTime_t(releaseDate).addYears(supportTime);
     }
 
@@ -576,8 +694,12 @@ int Package::state() const
 {
     int packageState = 0;
 
-    pkgDepCache::StateCache &stateCache = (*d->depCache)[*d->packageIter];
     const pkgCache::VerIterator &ver = d->packageIter->CurrentVer();
+    pkgDepCache::StateCache &stateCache = (*d->depCache)[*d->packageIter];
+
+    if (!d->staticStateCalculated) {
+        d->initStaticState(ver, stateCache);
+    }
 
     if (stateCache.Install()) {
         packageState |= ToInstall;
@@ -600,64 +722,24 @@ int Package::state() const
         packageState |= ToKeep;
     }
 
-    if (!ver.end()) {
-        packageState |= Installed;
-
-        if (stateCache.CandidateVer != NULL && stateCache.Upgradable()) {
-            packageState |= Upgradeable;
-            if (stateCache.Keep()) {
-                packageState |= Held;
-            }
-      }
-
-        if (stateCache.Downgrade()) {
-            packageState |= ToDowngrade;
-        }
-    } else {
-        packageState |= NotInstalled;
-    }
-    if (stateCache.NowBroken()) {
-        packageState |= NowBroken;
-    }
-
-    if (stateCache.InstBroken()) {
-        packageState |= InstallBroken;
-    }
-
-    if ((*d->packageIter)->Flags & (pkgCache::Flag::Important |
-                              pkgCache::Flag::Essential)) {
-        packageState |= IsImportant;
-    }
-
-    if ((*d->packageIter)->CurrentState == pkgCache::State::ConfigFiles) {
-        packageState |= ResidualConfig;
-    }
-
-    if (!stateCache.CandidateVer) {
-        packageState |= NotDownloadable;
-    } else if (!stateCache.CandidateVerIter(*d->depCache).Downloadable()) {
-        packageState |= NotDownloadable;
-    }
-
-    if (stateCache.Flags & pkgCache::Flag::Auto) {
-        packageState |= IsAuto;
-    }
-
-    if (stateCache.Garbage) {
-        packageState |= IsGarbage;
-    }
-
-    if (stateCache.NowPolicyBroken()) {
-        packageState |= NowPolicyBroken;
-    }
-
-    if (stateCache.InstPolicyBroken()) {
-        packageState |= InstallPolicyBroken;
-    }
-
    return packageState | d->state;
 }
 
+int Package::compareVersion(const QString &v1, const QString &v2)
+{
+    // Make deep copies of toStdString(), since otherwise they would
+    // go out of scope when we call c_str()
+    string s1 = v1.toStdString();
+    string s2 = v2.toStdString();
+
+    const char *a = s1.c_str();
+    const char *b = s2.c_str();
+    int lenA = strlen(a);
+    int lenB = strlen(b);
+
+    return _system->VS->DoCmpVersion(a, a+lenA, b, b+lenB);
+}
+
 bool Package::isInstalled() const
 {
     return !d->packageIter->CurrentVer().end();
@@ -676,6 +758,51 @@ bool Package::isSupported() const
     return false;
 }
 
+QList<DependencyItem> Package::depends() const
+{
+    return DependencyInfo::parseDepends(controlField("Depends"), Depends);
+}
+
+QList<DependencyItem> Package::preDepends() const
+{
+    return DependencyInfo::parseDepends(controlField("Pre-Depends"), PreDepends);
+}
+
+QList<DependencyItem> Package::suggests() const
+{
+    return DependencyInfo::parseDepends(controlField("Suggests"), Suggests);
+}
+
+QList<DependencyItem> Package::recommends() const
+{
+    return DependencyInfo::parseDepends(controlField("Recommends"), Recommends);
+}
+
+QList<DependencyItem> Package::conflicts() const
+{
+    return DependencyInfo::parseDepends(controlField("Conflicts"), Conflicts);
+}
+
+QList<DependencyItem> Package::replaces() const
+{
+    return DependencyInfo::parseDepends(controlField("Replaces"), Replaces);
+}
+
+QList<DependencyItem> Package::obsoletes() const
+{
+    return DependencyInfo::parseDepends(controlField("Obsoletes"), Obsoletes);
+}
+
+QList<DependencyItem> Package::breaks() const
+{
+    return DependencyInfo::parseDepends(controlField("Breaks"), Breaks);
+}
+
+QList<DependencyItem> Package::enhances() const
+{
+    return DependencyInfo::parseDepends(controlField("Enhance"), Enhances);
+}
+
 QStringList Package::dependencyList(bool useCandidateVersion) const
 {
     QStringList dependsList;
@@ -979,6 +1106,7 @@ bool Package::isTrusted() const
 
     pkgSourceList *Sources = d->backend->packageSourceList();
     QHash<pkgCache::PkgFileIterator, pkgIndexFile*> *trustCache = d->backend->cache()->trustCache();
+
     for (pkgCache::VerFileIterator i = Ver.FileList(); !i.end(); ++i)
     {
         pkgIndexFile *Index;
@@ -1113,4 +1241,9 @@ bool Package::setVersion(const QString &version)
     return true;
 }
 
+void Package::setPinned(bool pin)
+{
+    pin ? d->state |= IsPinned : d->state &= ~IsPinned;
+}
+
 }
