@@ -29,6 +29,7 @@
 
 #include <apt-pkg/pkgcache.h>
 
+#include "dependencyinfo.h"
 #include "globals.h"
 
 class pkgRecords;
@@ -116,6 +117,20 @@ public:
     */
     QString version() const;
 
+   /**
+    * Returns the upstream version of the package. This is the Debian version
+    * with the epoch and Debian revision information (if any) removed.
+    *
+    * (E.g. 0.2-0ubuntu1 becomes 0.2)
+    *
+    * @return The upstream version of the package
+    *
+    * @since 1.2
+    */
+    QString upstreamVersion() const;
+
+    static QString upstreamVersion(const QString &version);
+
    /**
     * Returns a list of all available versions of the package in the form of
     * "version, release" (E.g. "0.2-0ubuntu1, maverick")
@@ -230,6 +245,13 @@ public:
     */
     QString component() const;
 
+   /**
+    * Returns the md5sum of the candidate version of the package
+    *
+    * @return The md5sum of the package
+    */
+    QByteArray md5Sum() const;
+
    /**
     * Returns the url to the location of a package's changelog.
     *
@@ -323,6 +345,14 @@ public:
     */
     int state() const;
 
+   /**
+    * Compares v1 with v2 and returns an integer less than, equal to, or
+    * greater than zero if s1 is less than, equal to, or greater than s2.
+    *
+    * @since 1.2
+    */
+    static int compareVersion(const QString &v1, const QString &v2);
+
    /**
     * Returns whether the Package is installed
     */
@@ -333,6 +363,24 @@ public:
     */
     bool isSupported() const;
 
+    QList<DependencyItem> depends() const;
+
+    QList<DependencyItem> preDepends() const;
+
+    QList<DependencyItem> suggests() const;
+
+    QList<DependencyItem> recommends() const;
+
+    QList<DependencyItem> conflicts() const;
+
+    QList<DependencyItem> replaces() const;
+
+    QList<DependencyItem> obsoletes() const;
+
+    QList<DependencyItem> breaks() const;
+
+    QList<DependencyItem> enhances() const;
+
    /**
     * Returns a display-ready list of the names of all the dependencies of this package.
     *
@@ -448,7 +496,16 @@ public:
     */
     bool setVersion(const QString & version);
 
-    // "//" == TODO
+   /**
+    * Set the package as pinned internally for display purposes.
+    *
+    * To actually pin a package use @c Backend::setPackagePinned
+    *
+    * @since 1.2
+    * @see Backend::setPackagePinned()
+    */
+    void setPinned(bool pin);
+
    /**
     * An enumerator for various states that a @c Package may hold. A package
     * may hold several states at once.
@@ -503,7 +560,9 @@ public:
         /// The package's install policy is broken
         InstallPolicyBroken = 1 << 23,
         /// The package is not installed
-        NotInstalled        = 1 << 24
+        NotInstalled        = 1 << 24,
+        /// The package has been pinned
+        IsPinned            = 1 << 25
     };
     Q_DECLARE_FLAGS(States, State);
 
